"""
Schemas describe the different supported meta tag schemas and
the key names describing the attributes to expect.
For example, opengraph meta tags will contain the `og:name` tag
in a `property` attribute and the value in `content` while
`twitter` tags will be within a `name` attribute.

If the tag scheme is not supported it can be added in runtime
using the metatron.add_schema_spec
"""
SCHEMAS = {
    '': {
        'name': '',  # no schema
        'attribute': 'name',
        'value': 'content'
    },
    'og': {
        'name': 'og',
        'attribute': 'property',
        'value': 'content'
    },
    'twitter': {
        'name': 'twitter',
        'attribute': 'name',
        'value': 'content'
    }
}



from setuptools import setup, find_packages
from codecs import open
from os import path
from metatron.version import __version__

here = path.abspath(path.dirname(__file__))
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='metatron',
    version=__version__,
    description='Python 3 HTML meta tag parser, with emphasis on complex meta tag structures '
                'with support for OpenGraph and Twitter Card tags, including array handling',
    long_description=long_description,
    author='Harel Malka',
    author_email='harel@harelmalka.com',
    url='https://github.com/harel/metatron',
    download_url='https://github.com/harel/metatron/archive/0.1.tar.gz',
    keywords='html meta parser opengraph twittercard',
    install_requires=[
        'beautifulsoup4==4.6.0',
        'requests==2.18.4',
    ],
    license='MIT',
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Topic :: Communications :: Email',
        'Programming Language :: Python :: 3.5',
    ]
)



import unittest
from metatron import Metatron, add_schema_spec
from metatron.schemas import SCHEMAS


class TestMetatron(unittest.TestCase):
    def test_opengraph_nested(self):
        """
        {
            'image': {
                'value': 'http://example.com/ogp.jpg',
                'secure_url: 'https://secure.example.com/ogp.jpg',
            }
        }
        """
        content = """
            <meta name="description" content="Test Meta" />
            <meta property="someprop" content="Test Prop Value" />
            <meta property="og:title" content="The Two Image Shop" />
            <meta property="og:description" content="Its a shop with two images" />
            <meta property="og:image" content="http://example.com/ogp.jpg" />
            <meta property="og:image:secure_url" content="https://secure.example.com/ogp.jpg" />
            <meta property="og:image:type" content="image/jpeg" />
            <meta property="og:image:width" content="400" />
            <meta property="og:image:height" content="300" />
            <meta property="og:image:alt" content="A shiny red apple with a bite taken out" />
        """
        mt = Metatron(content=content, schemas=['og'])
        mt.traverse()
        assert mt['og']['image']['image'] == 'http://example.com/ogp.jpg'
        assert mt['og']['image']['secure_url'] == 'https://secure.example.com/ogp.jpg'


    def test_opengraph_nested_array(self):
        content = """
            <meta name="description" content="Test Meta" />
            <meta property="someprop" content="Test Prop Value" />
            <meta property="og:title" content="The Two Image Shop" />
            <meta property="og:title" content="The Two Image Store" />
            <meta property="og:description" content="Its a shop with two images" />
            <meta property="og:image" content="http://example.com/ogp.jpg" />
            <meta property="og:image:secure_url" content="https://secure.example.com/ogp.jpg" />
            <meta property="og:image:type" content="image/jpeg" />
            <meta property="og:image:width" content="400" />
            <meta property="og:image:height" content="300" />
            <meta property="og:image:alt" content="A shiny g apple with a bite taken out" />
            <meta property="og:image" content="http://example.com/ogp222.jpg" />
            <meta property="og:image:secure_url" content="https://secure.example.com/ogp.jpg" />
            <meta property="og:image:type" content="image/jpeg" />
            <meta property="og:image:width" content="500" />
            <meta property="og:image:height" content="600" />
            <meta property="og:image:alt" content="A shiny green apple with a bite taken out" />
        """
        mt = Metatron(content=content, schemas=['og'])
        mt.traverse()
        assert len(mt['og']['image']) == 2
        assert mt['og']['image'][0]['image'] == 'http://example.com/ogp.jpg'
        assert mt['og']['image'][0]['secure_url'] == 'https://secure.example.com/ogp.jpg'
        assert len(mt['og']['title']) == 2
        assert mt['og']['title'][1] == 'The Two Image Store'
        assert mt['og']['image'][1]['width'] == '500'

    def test_opengraph_array(self):
        content = """
            <meta name="description" content="Test Meta" />
            <meta property="someprop" content="Test Prop Value" />
            <meta property="og:image" content="http://example.com/ogp1.jpg" />
            <meta property="og:image:secure_url" content="https://secure.example.com/ogp1.jpg" />
            <meta property="og:image" content="http://example.com/ogp2.jpg" />
            <meta property="og:image:secure_url" content="https://secure.example.com/ogp2.jpg" />

        """
        mt = Metatron(content=content, schemas=['og'])
        mt.traverse()
        assert len(mt['og']['image']) == 2
        assert mt['og']['image'][0]['image'] == 'http://example.com/ogp1.jpg'


    def test_twitter_card_and_opengraph(self):
        """
        Collect both opengraph and twitter card schemas
        """
        content = """
            <meta property="og:title" content="Home - BBC News">
            <meta property="og:type" content="website">
            <meta property="og:description" content="Visit BBC News for up-to-the-minute news, breaking news, video, audio and feature stories. BBC News provides trusted World and UK news as well as local and regional perspectives. Also entertainment, business, science, technology and health news.">
            <meta property="og:site_name" content="BBC News">
            <meta property="og:locale" content="en_GB">
            <meta property="og:article:author" content="https://www.facebook.com/bbcnews">
            <meta property="og:article:section" content="Home">
            <meta property="og:url" content="http://www.bbc.co.uk/news">
            <meta property="og:image" content="//m.files.bbci.co.uk/modules/bbc-morph-news-waf-page-meta/2.1.0/bbc_news_logo.png">

            <meta name="twitter:card" content="summary_large_image">
            <meta name="twitter:site" content="@BBCNews">
            <meta name="twitter:title" content="Home - BBC News">
            <meta name="twitter:description" content="Visit BBC News for up-to-the-minute news, breaking news, video, audio and feature stories. BBC News provides trusted World and UK news as well as local and regional perspectives. Also entertainment, business, science, technology and health news.">
            <meta name="twitter:creator" content="@BBCNews">
            <meta name="twitter:image:src" content="//m.files.bbci.co.uk/modules/bbc-morph-news-waf-page-meta/2.1.0/bbc_news_logo.png">
            <meta name="twitter:image:alt" content="BBC News">
            <meta name="twitter:domain" content="www.bbc.co.uk">

            <link rel="apple-touch-icon-precomposed" sizes="57x57" href="//m.files.bbci.co.uk/modules/bbc-morph-news-waf-page-meta/2.1.0/apple-touch-icon-57x57-precomposed.png">
            <link rel="apple-touch-icon-precomposed" sizes="72x72" href="//m.files.bbci.co.uk/modules/bbc-morph-news-waf-page-meta/2.1.0/apple-touch-icon-72x72-precomposed.png">
            <link rel="apple-touch-icon-precomposed" sizes="114x114" href="//m.files.bbci.co.uk/modules/bbc-morph-news-waf-page-meta/2.1.0/apple-touch-icon-114x114-precomposed.png">
            <link rel="apple-touch-icon-precomposed" sizes="144x144" href="//m.files.bbci.co.uk/modules/bbc-morph-news-waf-page-meta/2.1.0/apple-touch-icon.png">
            <link rel="apple-touch-icon" href="//m.files.bbci.co.uk/modules/bbc-morph-news-waf-page-meta/2.1.0/apple-touch-icon.png">
            <meta name="msapplication-TileImage" content="//m.files.bbci.co.uk/modules/bbc-morph-news-waf-page-meta/2.1.0/windows-eight-icon-144x144.png">

            <meta name="theme-color" content="#bb1919">
            <meta name="msapplication-TileColor" content="#bb1919">

            <meta name="x-country" content="gb">
            <meta name="x-audience" content="Domestic">
            <meta name="CPS_AUDIENCE" content="Domestic">
            <meta name="CPS_CHANGEQUEUEID" content="115204091">
        """
        mt = Metatron(content=content, schemas=['og', 'twitter'])
        mt.traverse()
        assert 'twitter' in mt
        assert 'og' in mt
        assert mt['og']['image'] == '//m.files.bbci.co.uk/modules/bbc-morph-news-waf-page-meta/2.1.0/bbc_news_logo.png'
        assert mt['twitter']['domain'] == 'www.bbc.co.uk'
        assert mt['twitter']['description'].startswith('Visit BBC News')

    def test_simple_meta(self):
        """
        Test that simple meta tags are collected when no schema is provided
        """
        content = """
            <meta name="theme-color" content="#bb1919">
            <meta name="msapplication-TileColor" content="#bb1919">

            <meta name="x-country" content="gb">
            <meta name="x-audience" content="Domestic">
            <meta name="CPS_AUDIENCE" content="Domestic">
            <meta name="CPS_CHANGEQUEUEID" content="115204091">
        """
        mt = Metatron(content=content)
        mt.traverse()
        assert mt['']['CPS_CHANGEQUEUEID'] == '115204091'
        assert mt['']['x-country'] == 'gb'

    def test_add_schema(self):
        schema = {
            'name': 'test',
            'attribute': 'test',
            'value': 'test'
        }
        add_schema_spec(schema)
        assert 'test' in SCHEMAS
        assert SCHEMAS['test']['name'] == 'test'



if __name__ == '__main__':
    unittest.main()



"""
Read all html meta tags that comply with a scheme
(i.e., og: for opengraph, twitter, govuk etc.).
"""
from __future__ import absolute_import
import re
import requests
from sys import argv
from bs4 import BeautifulSoup
from functools import singledispatch
from metatron.schemas import SCHEMAS
from pprint import pprint


# a single dispatch to normalise schema argument into a list
@singledispatch
def _schemas(schemas):
    return schemas


@_schemas.register(str) # noqa
def _(schemas):
    return [schemas]


def add_schema_spec(spec):
    """
    Register a custom schema
    """
    if not validate_schema_spec(spec):
        raise Exception('Invalid schema spec. name, attribute and value keys are required')
    spec_tag = spec.get('name')
    SCHEMAS[spec_tag] = spec


def validate_schema_spec(spec):
    return all([key in spec for key in ['name', 'attribute', 'value']])


class Metatron(dict):
    def __init__(self, url=None, content=None, schemas=None, schema_spec=None, **kwargs):
        if not url and not content:
            raise Exception('Either url or content are required')
        self.url = url
        self.content = content
        self.schemas = schemas or ''
        self.last_tag = None
        if schema_spec:
            add_schema_spec(schema_spec)
        self.schemas = _schemas(self.schemas)
        super().__init__(**kwargs)

    def get_content(self):
        """
        Retrieve the content, either by fetching a url content provided
        as input
        """
        if self.url:
            response = requests.get(self.url)
            self.content = response.text
        return self.content

    def get_schema_regex(self, schema):
        """
        Schema regular expression for opengraph (og:) or
        twitter cards, etc.
        """
        if schema:
            return r'{0}:([\w:]+)'.format(schema)
        else:
            return r'([\w]+[^:])'

    def fetch_meta_tags(self):
        """
        Fetch meta tags from the source url or provided content
        """
        content = self.get_content()
        soup = BeautifulSoup(content, 'html.parser')
        tags = {}
        for schema in self.schemas:
            _spec = SCHEMAS[schema]
            schema_regex = self.get_schema_regex(schema)
            find_attrs = {_spec['attribute']: re.compile(schema_regex)}
            tags[schema] = soup.find_all('meta', attrs=find_attrs)
        return tags

    def name_path(self, tag, spec):
        """
        Extract the path, minus the schema, as a list:
        From:
            <meta property="og:image:secure_url" .../>
        To:
            ['image', 'secure_url']
        """
        meta_name = tag[spec['attribute']].split(spec.get('separator', ':'))
        if spec.get('name', '') in meta_name:
            meta_name.remove(spec.get('name', ''))
        return meta_name

    def tag_parts(self, tag, spec):
        meta_name = self.name_path(tag, spec)
        root = meta_name[0] if len(meta_name) > 1 else None
        return meta_name, root

    def next_tag(self, tags, current_index, spec):
        """
        Return the next tag, if available, and determine it's parts,
        root name and if it is indeed a root element
        """
        new_root = False
        try:
            current_tag = self.name_path(tags[current_index], spec)
            next_tag = tags[current_index + 1]
            next_parts, next_root = self.tag_parts(next_tag, spec)
            if next_root == current_tag[0] and len(next_parts) != len(current_tag):
                new_root = True
            return next_parts, next_root, new_root
        except (IndexError, TypeError):
            return None, None, new_root

    def key_value_root(self, tag, spec, separator=None):
        """
        For a given tag/spec, return the key, its value
        and the root element if available
        """
        separator = separator or ":"
        meta_name, root = self.tag_parts(tag, spec)
        if root:
            meta_name.remove(root)
        tag_value = tag[spec['value']]
        tag_key = separator.join(meta_name)
        return tag_key, tag_value, root

    def traverse(self):
        """
        traverse the tags and generae an index of each schema
        being fetched
        """
        meta_tags = self.fetch_meta_tags()
        meta = dict(self)
        for schema, tags in meta_tags.items():
            _index = {}
            meta.setdefault(schema, {})
            _spec = SCHEMAS[schema]
            for i, tag in enumerate(tags):
                key, value, root = self.key_value_root(tag, _spec)
                next_tag, next_root, new_root = self.next_tag(tags, i, _spec)
                if new_root and next_root:
                    if next_root in _index:
                        _index[next_root].append({key: value})
                    else:
                        _index[next_root] = [{key: value}]
                elif not root and key in _index:
                    _index[key].append({key: value})
                else:
                    _key = root or key
                    _index.setdefault(_key, [{}])
                    _index[_key][len(_index[_key]) - 1].update({key: value})
            self.reduce(schema, _index)

    def reduce(self, schema, index):
        """
        Reduce the index into a more compact structure:
            list of length 1 updates the top level schema dict
                if the value is not a dict
                otherwise set the root to that value
            list of length over 1 sets the value of the schema dict
                as reduced strings if the length of each dict is 1
                as is otherwise
        """
        self.setdefault(schema, {})
        for key, value in index.items():
            if len(value) == 1:
                if isinstance(value[0], dict) and len(list(value[0].keys())) > 1:
                    self[schema][key] = value[0]
                else:
                    self[schema].update(value[0])
            elif len(value) > 1:
                if all([len(list(item.keys())) == 1 for item in value]):
                    self[schema][key] = [item[key] for item in value]
                else:
                    self[schema][key] = value


if __name__ == '__main__':
    arg_count = len(argv)
    if arg_count > 1:
        url = argv[1]
        schemas = argv[2] if arg_count > 2 else []
        print("Getting: {0} (schema: {1})".format(url, schemas))
        mt = Metatron(url=url, schemas=schemas)
        mt.traverse()
        pprint(mt)
